from django.contrib import admin

from .models import APICredential, JWTSecret


@admin.register(JWTSecret)
class JWTSecretAdmin(admin.ModelAdmin):
    list_display = ('identifier',)
    search_fields = ('identifier',)


@admin.register(APICredential)
class APICredentialAdmin(admin.ModelAdmin):
    list_display = ('api_root', 'client_id')
    search_fields = ('api_root',)



from urllib.parse import urlsplit, urlunsplit

from django.db import models
from django.db.models.functions import Length
from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse
from zds_client import ClientAuth


class APIMixin:

    def get_absolute_api_url(self, request=None, **kwargs) -> str:
        """
        Build the absolute URL of the object in the API.
        """
        # build the URL of the informatieobject
        resource_name = self._meta.model_name

        reverse_kwargs = {'uuid': self.uuid}
        reverse_kwargs.update(**kwargs)

        url = reverse(
            f'{resource_name}-detail',
            kwargs=reverse_kwargs, request=request
        )
        return url


class JWTSecret(models.Model):
    identifier = models.CharField(_("identifier"), max_length=50, unique=True)
    secret = models.CharField(_("secret"), max_length=255)

    def __str__(self):
        return self.identifier


class APICredential(models.Model):
    """
    Store credentials for external APIs.

    When we need to authenticate against a remote API, we need to know which
    client ID and secret to use to sign the JWT.
    """
    api_root = models.URLField(_("api root"), unique=True)
    client_id = models.CharField(_("client id"), max_length=255)
    secret = models.CharField(_("secret"), max_length=255)

    class Meta:
        verbose_name = _("API credential")
        verbose_name_plural = _("API credentials")

    def __str__(self):
        return self.api_root

    @classmethod
    def get_auth(cls, url: str, **kwargs) -> ClientAuth:
        split_url = urlsplit(url)
        scheme_and_domain = urlunsplit(split_url[:2] + ('', '', ''))

        candidates = (
            cls.objects
            .filter(api_root__startswith=scheme_and_domain)
            .annotate(api_root_length=Length('api_root'))
            .order_by('-api_root_length')
        )

        # select the one matching
        for candidate in candidates.iterator():
            if url.startswith(candidate.api_root):
                credentials = candidate
                break
        else:
            return None

        auth = ClientAuth(
            client_id=credentials.client_id,
            secret=credentials.secret,
            **kwargs
        )
        return auth



__all__ = [
    'API_VERSION',
    'BASE_REST_FRAMEWORK',
    'BASE_SWAGGER_SETTINGS',
    'LINK_FETCHER',
    'ZDS_CLIENT_CLASS',
    'GEMMA_URL_TEMPLATE',
    'GEMMA_URL_COMPONENTTYPE',
    'GEMMA_URL_INFORMATIEMODEL',
    'GEMMA_URL_INFORMATIEMODEL_VERSIE',
]

API_VERSION = '1.0.0-alpha'  # semantic version

BASE_REST_FRAMEWORK = {
    'DEFAULT_RENDERER_CLASSES': (
        'djangorestframework_camel_case.render.CamelCaseJSONRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        'djangorestframework_camel_case.parser.CamelCaseJSONParser',
    ),
    # 'DEFAULT_AUTHENTICATION_CLASSES': (
    #     'oauth2_provider.contrib.rest_framework.OAuth2Authentication',
    #     # 'rest_framework.authentication.SessionAuthentication',
    #     # 'rest_framework.authentication.BasicAuthentication'
    # ),
    # 'DEFAULT_PERMISSION_CLASSES': (
    #     'oauth2_provider.contrib.rest_framework.TokenHasReadWriteScope',
    #     # 'rest_framework.permissions.IsAuthenticated',
    #     # 'rest_framework.permissions.AllowAny',
    # ),
    'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.URLPathVersioning',
    #
    # # Generic view behavior
    # 'DEFAULT_PAGINATION_CLASS': 'ztc.api.utils.pagination.HALPagination',
    'DEFAULT_FILTER_BACKENDS': (
        'zds_schema.filters.Backend',
        # 'rest_framework.filters.SearchFilter',
        # 'rest_framework.filters.OrderingFilter',
    ),
    #
    # # Filtering
    # 'SEARCH_PARAM': 'zoek',  # 'search',
    # 'ORDERING_PARAM': 'sorteer',  # 'ordering',
    #
    # Versioning
    'DEFAULT_VERSION': '1',  # NOT to be confused with API_VERSION - it's the major version part
    'ALLOWED_VERSIONS': ('1', ),
    'VERSION_PARAM': 'version',
    #
    # # Exception handling
    'EXCEPTION_HANDLER': 'zds_schema.views.exception_handler',
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
}

BASE_SWAGGER_SETTINGS = {
    # 'SECURITY_DEFINITIONS': {
    #     'OAuth2': {
    #         'type': 'oauth2',
    #         'flow': 'application',
    #         'tokenUrl': '/oauth2/token/',
    #         'scopes': {
    #             'write': 'Schrijftoegang tot de catalogus en gerelateerde objecten.',
    #             'read': 'Leestoegang tot de catalogus en gerelateerde objecten.'
    #         }
    #     },
    #     'Bearer': {
    #         'type': 'apiKey',
    #         'name': 'Authorization',
    #         'in': 'header'
    #     },
    # },
    'DEFAULT_AUTO_SCHEMA_CLASS': 'zds_schema.inspectors.view.AutoSchema',
    'DEFAULT_INFO': 'must.be.overridden',
    'DEFAULT_FIELD_INSPECTORS': (
        'zds_schema.inspectors.geojson.GeometryFieldInspector',

        'drf_yasg.inspectors.CamelCaseJSONFilter',
        'drf_yasg.inspectors.RecursiveFieldInspector',
        'drf_yasg.inspectors.ReferencingSerializerInspector',
        'drf_yasg.inspectors.ChoiceFieldInspector',
        'drf_yasg.inspectors.FileFieldInspector',
        'drf_yasg.inspectors.DictFieldInspector',
        'drf_yasg.inspectors.HiddenFieldInspector',
        'drf_yasg.inspectors.RelatedFieldInspector',
        'drf_yasg.inspectors.SimpleFieldInspector',
        'drf_yasg.inspectors.StringDefaultFieldInspector',
    ),
    'DEFAULT_FILTER_INSPECTORS': (
        'zds_schema.inspectors.query.FilterInspector',
    )
}

REDOC_SETTINGS = {
    'EXPAND_RESPONSES': '200,201'
}

# See: https://github.com/Rebilly/ReDoc#redoc-options-object
LINK_FETCHER = 'requests.get'

ZDS_CLIENT_CLASS = 'zds_client.Client'

GEMMA_URL_TEMPLATE = "https://www.gemmaonline.nl/index.php/{informatiemodel}_{versie}/doc/{componenttype}/{component}"
GEMMA_URL_COMPONENTTYPE = 'objecttype'
GEMMA_URL_INFORMATIEMODEL = 'Rgbz'
GEMMA_URL_INFORMATIEMODEL_VERSIE = '2.0'



from django.utils.translation import ugettext_lazy as _

from rest_framework import status
from rest_framework.exceptions import APIException


class Conflict(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = _("A conflict occurred")
    default_code = 'conflict'


class Gone(APIException):
    status_code = status.HTTP_410_GONE
    default_detail = _("The resource is gone")
    default_code = 'gone'


class PreconditionFailed(APIException):
    status_code = status.HTTP_412_PRECONDITION_FAILED
    default_detail = _("Precondition failed")
    default_code = 'precondition_failed'



# https://pyjwt.readthedocs.io/en/latest/usage.html#reading-headers-without-validation
# -> we can put the organization/service in the headers itself
import logging
from typing import Union

from django.conf import settings
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _

import jwt
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response

from .constants import VERSION_HEADER
from .models import JWTSecret
from .scopes import Scope

logger = logging.getLogger(__name__)

EMPTY_PAYLOAD = {
    'scopes': [],
    'zaaktypes': [],
}


class JWTPayload:

    def __init__(self, encoded: str=None):
        self.encoded = encoded

    def __repr__(self):
        return "<%s: payload=%r>" % (self.__class__.__name__, self.payload)

    def __getitem__(self, key):
        return self.payload[key]

    def get(self, attr, *args, **kwargs):
        return self.payload.get(attr, *args, **kwargs)

    @cached_property
    def payload(self) -> dict:
        if self.encoded is None:
            return EMPTY_PAYLOAD

        try:
            header = jwt.get_unverified_header(self.encoded)
        except jwt.DecodeError:
            logger.info("Invalid JWT encountered")
            raise PermissionDenied(
                _('JWT could not be decoded. Possibly you made a copy-paste mistake.'),
                code='jwt-decode-error'
            )

        try:
            jwt_secret = JWTSecret.objects.get(identifier=header['client_identifier'])
        except (JWTSecret.DoesNotExist, KeyError):
            raise PermissionDenied(
                'Client credentials zijn niet aanwezig',
                code='missing-client-identifier'
            )
        else:
            key = jwt_secret.secret

        # the jwt package does verification against tampering (TODO: unit test)
        try:
            payload = jwt.decode(self.encoded, key, algorithms='HS256')
        except jwt.InvalidSignatureError as exc:
            logger.exception("Invalid signature - possible payload tampering?")
            raise PermissionDenied(
                'Client credentials zijn niet geldig',
                code='invalid-jwt-signature'
            )

        return payload.get('zds', EMPTY_PAYLOAD)

    def has_scopes(self, scopes: Union[Scope, None]) -> bool:
        """
        Check whether all of the expected scopes are present or not.
        """
        if scopes is None:
            # TODO: should block instead of allow it - we'll gradually introduce this
            return True

        scopes_provided = self.payload['scopes']
        # simple form - needs a more complex setup if a scope 'bundles'
        # other scopes
        return scopes.is_contained_in(scopes_provided)


class AuthMiddleware:

    header = 'HTTP_AUTHORIZATION'
    auth_type = 'Bearer'

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        self.extract_jwt_payload(request)
        return self.get_response(request) if self.get_response else None

    def extract_jwt_payload(self, request):
        authorization = request.META.get(self.header, '')
        prefix = f"{self.auth_type} "
        if authorization.startswith(prefix):
            # grab the actual token
            encoded = authorization[len(prefix):]
        else:
            encoded = None

        request.jwt_payload = JWTPayload(encoded)


class APIVersionHeaderMiddleware:
    """
    Include a header specifying the API-version
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if self.get_response is None:
            return None

        response = self.get_response(request)

        # not an API response, exit early
        if not isinstance(response, Response):
            return response

        # set the header
        response[VERSION_HEADER] = settings.API_VERSION

        return response



"""
Patch the content-type of error responses.

Due to the changes between Swagger 2.0 and OpenAPI 3.0, we cannot handle
this at the Python level.
"""
from django.core.management import BaseCommand

import oyaml as yaml

from ...views import ERROR_CONTENT_TYPE


class Command(BaseCommand):
    help = "Patch the error-response content types in the OAS 3 spec"

    def add_arguments(self, parser):
        parser.add_argument('api-spec', help="Path to the openapi spec. Will be overwritten!")

    def handle(self, **options):
        source = options['api-spec']
        with open(source, 'r') as infile:
            spec = yaml.safe_load(infile)

        for path, methods in spec['paths'].items():
            for method in methods.values():
                if 'responses' not in method:
                    continue

                for status, response in method['responses'].items():
                    if not (400 <= int(status) < 600):
                        continue

                    content = {}
                    for contenttype, _response in response['content'].items():
                        if contenttype == 'application/json':
                            contenttype = ERROR_CONTENT_TYPE
                        content[contenttype] = _response

                    response['content'] = content

        with open(source, 'w') as outfile:
            yaml.dump(spec, outfile, default_flow_style=False)



from django.utils.translation import ugettext_lazy as _

from rest_framework.exceptions import ValidationError
from rest_framework.settings import api_settings

from .filters import Backend
from .utils import lookup_kwargs_to_filters, underscore_to_camel


class NestedViewSetMixin:
    def get_queryset(self):
        """
        Filter the ``QuerySet`` based on its parents.
        """
        queryset = super().get_queryset()
        serializer_class = self.get_serializer_class()

        lookup_kwargs = getattr(serializer_class, 'parent_lookup_kwargs', {})
        filters = lookup_kwargs_to_filters(lookup_kwargs, self.kwargs)

        return queryset.filter(**filters)


class CheckQueryParamsMixin:

    def _check_query_params(self, request) -> None:
        """
        Validate that the query params in the request are known.
        """
        # nothing to check if there are no query parameters
        if not request.query_params:
            return

        # TODO: check for pagination params
        # NOTE: only works with django_filters based filter backends
        backend = Backend()
        queryset = self.get_queryset()
        filterset_class = backend.get_filterset_class(self, queryset)

        # build a list of known params from the filters
        filters = filterset_class().get_filters().keys()

        known_params = {underscore_to_camel(param) for param in filters}

        unknown_params = set(request.query_params.keys()) - known_params
        if unknown_params:
            msg = _("Onbekende query parameters: %s" % ", ".join(unknown_params))
            raise ValidationError(
                {api_settings.NON_FIELD_ERRORS_KEY: msg},
                code='unknown-parameters'
            )

    def list(self, request, *args, **kwargs):
        self._check_query_params(request)
        return super().list(request, *args, **kwargs)



from collections import OrderedDict

from drf_yasg import openapi
from drf_yasg.inspectors import FieldInspector, NotHandled
from rest_framework import serializers
from rest_framework_gis.fields import GeometryField

from ..geo import DEFAULT_CRS, HEADER_ACCEPT, HEADER_CONTENT

REF_NAME_GEOJSON_GEOMETRY = 'GeoJSONGeometry'


def register_geojson(definitions):
    Geometry = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        title="Geometry",
        description="GeoJSON geometry",
        discriminator='type',
        required=['type'],
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1'),
        properties=OrderedDict((
            ('type', openapi.Schema(
                type=openapi.TYPE_STRING,
                enum=[
                    'Point',
                    'MultiPoint',
                    'LineString',
                    'MultiLineString',
                    'Polygon',
                    'MultiPolygon',
                    'Feature',
                    'FeatureCollection',
                    'GeometryCollection',
                ],
                description="The geometry type"
            )),
        ))
    )
    definitions.set('Geometry', Geometry)

    Point2D = openapi.Schema(
        type=openapi.TYPE_ARRAY,
        title="Point2D",
        description="A 2D point",
        items=openapi.Schema(type=openapi.TYPE_NUMBER),
        maxItems=2,
        minItems=2
    )
    definitions.set('Point2D', Point2D)

    Point = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON point geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.2'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.SchemaRef(definitions, 'Point2D')),
                ))
            )
        ]
    )
    definitions.set('Point', Point)

    MultiPoint = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-point geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.3'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.SchemaRef(definitions, 'Point2D'),
                    )),
                ))
            )
        ]
    )
    definitions.set('MultiPoint', MultiPoint)

    LineString = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON line-string geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.4'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.SchemaRef(definitions, 'Point2D'),
                        minItems=2,
                    )),
                ))
            )
        ]
    )
    definitions.set('LineString', LineString)

    MultiLineString = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-line-string geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.5'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.Schema(
                            type=openapi.TYPE_ARRAY,
                            items=openapi.SchemaRef(definitions, 'Point2D')
                        )
                    )),
                ))
            )
        ]
    )
    definitions.set('MultiLineString', MultiLineString)

    Polygon = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON polygon geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.6'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.Schema(
                            type=openapi.TYPE_ARRAY,
                            items=openapi.SchemaRef(definitions, 'Point2D')
                        )
                    )),
                ))
            )
        ]
    )
    definitions.set('Polygon', Polygon)

    MultiPolygon = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-polygon geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.7'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['coordinates'],
                properties=OrderedDict((
                    ('coordinates', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.Schema(
                            type=openapi.TYPE_ARRAY,
                            items=openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.SchemaRef(definitions, 'Point2D')
                            )
                        )
                    )),
                ))
            )
        ]
    )
    definitions.set('MultiPolygon', MultiPolygon)

    GeometryCollection = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-polygon geometry",
        externalDocs=OrderedDict(url='https://tools.ietf.org/html/rfc7946#section-3.1.8'),
        allOf=[
            openapi.SchemaRef(definitions, 'Geometry'),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=['geometries'],
                properties=OrderedDict((
                    ('geometries', openapi.Schema(
                        type=openapi.TYPE_ARRAY,
                        items=openapi.SchemaRef(definitions, 'Geometry')
                    )),
                ))
            )
        ]
    )
    definitions.set('GeometryCollection', GeometryCollection)

    GeoJSONGeometry = openapi.Schema(
        title=REF_NAME_GEOJSON_GEOMETRY,
        type=openapi.TYPE_OBJECT,
        oneOf=[
            openapi.SchemaRef(definitions, 'Point'),
            openapi.SchemaRef(definitions, 'MultiPoint'),
            openapi.SchemaRef(definitions, 'LineString'),
            openapi.SchemaRef(definitions, 'MultiLineString'),
            openapi.SchemaRef(definitions, 'Polygon'),
            openapi.SchemaRef(definitions, 'MultiPolygon'),
            openapi.SchemaRef(definitions, 'GeometryCollection'),
        ]
    )
    definitions.set(REF_NAME_GEOJSON_GEOMETRY, GeoJSONGeometry)


class GeometryFieldInspector(FieldInspector):

    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kwargs):
        if not isinstance(field, GeometryField):
            return NotHandled

        definitions = self.components.with_scope(openapi.SCHEMA_DEFINITIONS)

        if not definitions.has('Geometry'):
            register_geojson(definitions)

        return openapi.SchemaRef(definitions, REF_NAME_GEOJSON_GEOMETRY)

    def has_geo_fields(self, serializer) -> bool:
        """
        Check if any of the serializer fields are a GeometryField.

        If the serializer has nested serializers, a depth-first search is done
        to check if the nested serializers has `GeometryField`\ s.
        """
        for field in serializer.fields.values():
            if isinstance(field, serializers.Serializer):
                has_nested_geo_fields = self.probe_inspectors(
                    self.field_inspectors, 'has_geo_fields',
                    field, {'field_inspectors': self.field_inspectors}
                )
                if has_nested_geo_fields:
                    return True

            elif isinstance(field, (serializers.ListSerializer, serializers.ListField)):
                field = field.child

            if isinstance(field, GeometryField):
                return True

        return False

    def get_request_header_parameters(self, serializer):
        if not self.has_geo_fields(serializer):
            return []

        # see also http://lyzidiamond.com/posts/4326-vs-3857 for difference
        # between coordinate system and projected coordinate system
        return [openapi.Parameter(
            name=HEADER_ACCEPT,
            type=openapi.TYPE_STRING,
            in_=openapi.IN_HEADER,
            required=True,
            description="Het 'Coordinate Reference System' (CRS) van de "
                        "verzoekdata. Volgens de GeoJSON spec is WGS84 de "
                        "default (EPSG:4326 is hetzelfde als WGS84).",
            enum=[DEFAULT_CRS]
        )]

    def get_response_headers(self, serializer, status=None):
        if not self.has_geo_fields(serializer):
            return None

        if int(status) != 200:
            return None

        return OrderedDict((
            (HEADER_CONTENT, openapi.Schema(
                type=openapi.TYPE_STRING,
                enum=[DEFAULT_CRS],
                description="Het 'Coordinate Reference System' (CRS) van de "
                            "antwoorddata. Volgens de GeoJSON spec is WGS84 de "
                            "default (EPSG:4326 is hetzelfde als WGS84).",
            )),
        ))
